// AlignmentMonitoring
#include "AlignmentMonitoring/CompareConstants.h"
#include "AlignmentMonitoring/ParseConstants.h"
#include "AlignmentMonitoring/Utilities.h"
// STL
#include <algorithm>
#include <iostream>
#include <fstream>
#include <sstream>

// Boost
#include "boost/regex.hpp"
// ROOT
#include "TFile.h"


#include "GaudiKernel/SmartDataPtr.h"
#include "GaudiKernel/MsgStream.h"
#include "GaudiKernel/IDataProviderSvc.h"
#include "GaudiKernel/IToolSvc.h"

using namespace Alignment::AlignmentMonitoring;

// Constructors
CompareConstants::CompareConstants()
  : CompareConstants("","","")
{
}

CompareConstants::CompareConstants(std::map<std::string,double> iVals, std::map<std::string,double> fVals)
  : CompareConstants("","","")
{
  m_constRef = iVals;
  m_constAlt = fVals;
}

CompareConstants::CompareConstants(const char* system, const char* ref, const char* dir)
: m_verbose(false)
, m_system(system)
, m_ref(ref)
, m_dir(dir)
{
  m_dofs.push_back("Tx");
  m_dofs.push_back("Ty");
  m_dofs.push_back("Tz");
  m_dofs.push_back("Rx");
  m_dofs.push_back("Ry");
  m_dofs.push_back("Rz");
  SetRanges();
}

// Methods
void
CompareConstants::Compare()
{
  if (m_verbose) std::cout  << "nElements: " << m_constRef.size() << " " << m_constAlt.size() << std::endl;
  if (m_constAlt.size() != m_constRef.size()) {
    std::cout << "Beware, the number of elements for the two maps is different! (" << m_constAlt.size() << " vs " << m_constRef.size() << ") \n";
    return;
  }
  for (auto iel : m_constRef) {
    m_mapWarnings[iel.first] = CheckConstant(iel.first.c_str(), iel.second - m_constAlt[iel.first]);
    if (m_verbose)
      std::cout << iel.first << ": "
                << iel.second << " - " << m_constAlt[iel.first] << " = " << iel.second - m_constAlt[iel.first]
                << " : " << m_mapWarnings[iel.first]
                << std::endl;
    FillDeltaHist(iel.first.c_str());
  }
  return;
}

void
CompareConstants::Compare(const char* ver)
{
  std::map<std::string, WarningLevel> wls;
  ParseConstants parser;
  m_constRef = parser.Parse((m_dir+"/"+m_system+"/"+m_ref+".xml").c_str());
  m_constAlt = parser.Parse((m_dir+"/"+m_system+"/"+std::string(ver)+".xml").c_str());
  Compare();
  return;
}

void
CompareConstants::Compare(const char* xml1, const char* xml2)
{
  std::map<std::string, WarningLevel> wls;
  ParseConstants parser;
  m_constRef = parser.Parse(xml1);
  m_constAlt = parser.Parse(xml2);
  Compare();
  return;
}

void
CompareConstants::Compare(const std::vector< std::string > xmls1, const std::vector< std::string > xmls2)
{
  std::map<std::string, WarningLevel> wls;
  ParseConstants parser;
  m_constRef = parser.Parse(xmls1);
  m_constAlt = parser.Parse(xmls2);
  Compare();
}

void CompareConstants::CompareWithRun(const char* dir, const char* runNo){
  std::map<std::string, WarningLevel> wls;
  ParseConstants parser;
  m_constRef = parser.Parse((m_dir+"/"+m_system+"/"+m_ref+".xml").c_str());
  //std::cout << "reference path: " << m_dir+"/"+m_system+"/"+m_ref+".xml"  << std::endl;
  m_constAlt = parser.Parse((std::string(dir)+"/"+std::string(runNo)+"/xml/"+m_system+".xml").c_str());
  //std::cout << "new path: " << std::string(dir)+"/"+std::string(runNo)+"/xml/"+m_system+".xml" << std::endl;
  if (m_verbose) std::cout  << "nElements: " << m_constRef.size() << " " << m_constAlt.size() << std::endl;
  if (m_constAlt.size() != m_constRef.size()) {
    std::cout << "Beware, the number of elements for the two files\n"
              << "\t" << m_dir+"/"+m_system+"/"+m_ref+".xml" << "\n"
              << "\t" << std::string(dir)+"/"+std::string(runNo)+"/xml/"+m_system+".xml" << "\n"
              << "is not the same. Please check your options!\n";
    return;
  }
  for (auto iel : m_constRef) {
    std::vector<double> limits = GetLimits(iel.first.c_str());
    m_mapWarnings[iel.first] = CheckConstant(iel.second - m_constAlt[iel.first], limits[0], limits[1]);
    if (m_verbose)
      std::cout << iel.first << ": "
                << iel.second << " - " << m_constAlt[iel.first] << " = " << iel.second - m_constAlt[iel.first]
                << " : " << m_mapWarnings[iel.first]
                << std::endl;
    FillDeltaHist(iel.first.c_str());
  }
  return;
}

WarningLevel
CompareConstants::CheckConstant(const char* name, int dof, double delta)
{
  return CheckConstant(name, m_dofs[dof].c_str(), delta);
}

WarningLevel
CompareConstants::CheckConstant(const char* name, const char* dof, double delta)
{
  return CheckConstant( ( std::string(name)+std::string(".")+std::string(dof) ).c_str(), delta );
}

WarningLevel
CompareConstants::CheckConstant(const char* name, double delta)
{
  std::vector<double> limits = GetLimits( name );
  WarningLevel wnl = CheckConstant(delta, limits[0], limits[1]);
  return wnl;
}

void
CompareConstants::PrintWarnings(int level, MsgStream& err)
{

  if (!CheckWarningLevel(level) ) return;
  
  std::string warning_string = "OK";


  for (auto wrn : m_mapWarnings) {
    //if (!m_verbose && wrn.second == WarningLevel::OK) continue;
    if (wrn.second < level) continue;
    if(wrn.second==WarningLevel::OK)      warning_string = "\033[1;32m OK \033[0m";
    if(wrn.second==WarningLevel::WARNING) warning_string = "\033[1;33m WARNING \033[0m";
    if(wrn.second==WarningLevel::SEVERE)  warning_string = "\033[1;31m SEVERE \033[0m";
  
    err << MSG::WARNING   << wrn.first << ": " << warning_string 
              << "\t(" << m_constAlt[wrn.first] << " - " << m_constRef[wrn.first]  << " = " << m_constAlt[wrn.first]-m_constRef[wrn.first] << ") "
    	      << "\t limits = ["<< GetLimits(wrn.first.c_str())[0] << "," << GetLimits(wrn.first.c_str())[1] <<"]"
	<< endmsg;
  }
  return;
}

void
CompareConstants::PrintWarnings(int level)
{

  if (!CheckWarningLevel(level) ) return;
  
  std::string warning_string = "OK";
  std::stringstream ss;


  for (auto wrn : m_mapWarnings) {
    //if (!m_verbose && wrn.second == WarningLevel::OK) continue;
    if (wrn.second < level) continue;
    if(wrn.second==WarningLevel::OK)      warning_string = "\033[1;32m OK \033[0m";
    if(wrn.second==WarningLevel::WARNING) warning_string = "\033[1;33m WARNING \033[0m";
    if(wrn.second==WarningLevel::SEVERE)  warning_string = "\033[1;31m SEVERE \033[0m";
 
    std::cout << wrn.first << ": " << warning_string 
              << "\t(" << m_constAlt[wrn.first] << " - " << m_constRef[wrn.first]  << " = " << m_constAlt[wrn.first]-m_constRef[wrn.first] << ") "
    	      << "\t limits = ["<< GetLimits(wrn.first.c_str())[0] << "," << GetLimits(wrn.first.c_str())[1] <<"]"
              << std::endl;
    std::cout.flush();
  }
  return;
}


std::string
CompareConstants::GetWarningMessages(int level) // return a string with the warnings
{

  if (!CheckWarningLevel(level) )
    {
      std::stringstream ss;
      ss << "Warning level can be 1, 2 or 3 Wrong warning level "<<level<<" given";
      return ss.str();
      
    }
  
  std::string warning_string = "OK";
  std::stringstream ss;


  for (auto wrn : m_mapWarnings) {
    //if (!m_verbose && wrn.second == WarningLevel::OK) continue;
    if (wrn.second != level) continue;
    if(wrn.second==WarningLevel::OK)      warning_string = " OK ";
    if(wrn.second==WarningLevel::WARNING) warning_string = " WARNING ";
    if(wrn.second==WarningLevel::SEVERE)  warning_string = " SEVERE ";
 
    ss << wrn.first << ": " << warning_string 
       << "\t(" << m_constAlt[wrn.first] << " - " << m_constRef[wrn.first]  << " = " << m_constAlt[wrn.first]-m_constRef[wrn.first] << ") "
       << "\t limits = ["<< GetLimits(wrn.first.c_str())[0] << "," << GetLimits(wrn.first.c_str())[1] <<"]"
       << std::endl;
  }
  return ss.str();
}



WarningLevel
CompareConstants::CheckConstant(double delta, double d1, double d2)
{
  if (m_verbose) std::cout << fabs(delta) << "\t["<< fabs(d1) << "," <<fabs(d2)<<"]:";
  WarningLevel wl = WarningLevel::SEVERE; // default=SEVERE
  if (fabs(delta) < fabs(d1)) wl = WarningLevel::OK; // OK within range
  else if (fabs(delta) < fabs(d2)) wl = WarningLevel::WARNING; // WARNING within double range
  if (m_verbose) std::cout << "\t" << wl << std::endl;
  return wl;
}

std::vector<double>
CompareConstants::GetLimits(const char* element)
{
  std::vector<double> limits;
  const std::string s(element);
  for (auto rng : m_rngs) {
    /* debug
    if (std::string(element).find("odule1B.")!= std::string::npos)
      std::cout << element << " " << rng.first << ": "
                << (std::string(element).find(rng.first.c_str()) != std::string::npos) << " "
                << (std::string(element).find("Module1B.") != std::string::npos) << " "
                << rng.first.compare("Module1B.") << std::endl;
     */
    //if (std::string(element).find(rng.first) != std::string::npos) { // needs to be updated with regular expressions (regex) // Old
    boost::regex rgx(rng.first);
    if (boost::regex_search(s.begin(), s.end(), rgx)) {
      limits = rng.second;
      if (m_verbose) std::cout<< "element: "<< element << " \tfound: " << rng.first << " (" << rng.second[0] << "," << rng.second[0] << ")\n";
      break;
    }
     
  }
  if (limits.size() < 1) {
    std::cout << "Limits for element " << element << " not found. Returning vector with 0's.";
    // limits = {0,0};
    limits = {0.0000015,0.000001};
  }
  return limits;
}

void
CompareConstants::FillDeltaHist(const char* element)
{
  Utilities ut;
  const std::string s(element);
  for (auto rng : m_rngs) {
    boost::regex rgx(rng.first);
    if (boost::regex_search(s.begin(), s.end(), rgx)) {
      double delta = m_constRef[element] - m_constAlt[element];
      if (m_verbose) std::cout<< "element: "<< element << " \tfound: " << rng.first << "\tdelta = " << delta << std::endl;
      std::string hname(rng.first);
      hname = ut.find_and_replace(hname, ".*", "");
      hname = ut.find_and_replace(hname, "\\.", "");
      if ( m_hists.find(hname) == m_hists.end() ) m_hists[hname] = new TH1D(hname.c_str(), hname.c_str(), 20, -rng.second[1], rng.second[1]);
      m_hists[hname]->Fill(delta);
      break;
    }
  }
  return;
}

void
CompareConstants::SaveHists(const char* filename)
{
  TFile f(filename,"RECREATE");
  f.cd();
  for (auto his : m_hists)
    his.second->Write();
  f.Close();
}

void
CompareConstants::SetRanges(const char* filename)
{
  Utilities ut;
  std::string line;
  std::ifstream ifile( std::string(filename).size() == 0 ? (std::getenv("ALIGNMENTMONITORINGROOT")+std::string("/files/ConstantsReferences.txt")).c_str() : filename);
  if (ifile.is_open()) {
    while ( std::getline (ifile,line) ) {
      std::vector<std::string> refs = ut.splitString(line," \t");
      if (refs.size() < 2 ||
          refs[0].find("#") != std::string::npos) continue;
      m_rngs[refs[0]] = {std::stod(refs[1]),std::stod(refs[2])};
    }
    ifile.close();
  }
  // if (1) //debug
  // std::cout<<"\nBegin elements checked\n";
  // for (auto coso : m_rngs)
  // std::cout << coso.first << "\t["<< coso.second[0] << "," << coso.second[1] <<"]:" << std::endl;
  // std::cout<<"End elements checked\n\n";
  // return;
}

bool
CompareConstants::CheckWarningLevel(int level)
{
  if (level< 1 || level > 3) {
    std::cout << "You selected a warning level that is not valid (1-3)!\n";
    return false;
  }
  return true;
}

int
CompareConstants::GetNumWarnings(int level)
{
  if ( !CheckWarningLevel(level) ) return -1;
  // get number of warnings of this level
  int nw(0);
  for (auto wrn : m_mapWarnings)
    if (wrn.second == level) ++nw;
  return nw;
}

int
CompareConstants::GetNumWarnings(int level, const char* regex)
{
  std::string s(regex);
  boost::regex rgx(s);
  if ( !CheckWarningLevel(level) ) return -1;
  // get number of warnings of this level
  int nw(0);
  for (auto wrn : m_mapWarnings){
    if (boost::regex_search(wrn.first.begin(), wrn.first.end(), rgx)) 
      if (wrn.second == level) ++nw;
  }
  return nw;
}


std::string
CompareConstants::GetAlignablesWithWarnings(int level)
{
  std::string str_alignabes;
  if ( !CheckWarningLevel(level) ) return str_alignabes;
  // get number of warnings of this level
  for (auto wrn : m_mapWarnings)
    if (wrn.second == level)  str_alignabes += std::string(wrn.first) + std::string(";");
  return str_alignabes;
}


std::string
CompareConstants::GetAlignablesWithWarnings(int level, const char* regex)
{
  std::string s(regex);
  boost::regex rgx(s);
  std::string str_alignabes;
  if ( !CheckWarningLevel(level) ) return str_alignabes;
  // get number of warnings of this level
  for (auto wrn : m_mapWarnings)
    if (boost::regex_search(wrn.first.begin(), wrn.first.end(), rgx)) 
      if (wrn.second == level) str_alignabes += std::string(wrn.first) + std::string(";");
  return str_alignabes;
}
